:- module(corrector, [corrector/3]).

:- use_module(store,[
        pg_ShowSTATE/3,
        proc_DeleteCLS/3,
        pg_CLAUSETEMPLATE/7,
        proc_MODIFY/2,
        proc_ReadCLS/2,
        proc_CLS/3,
        pg_PROC/3,
        proc_SHOW/1,
        cls_MODIFY/2,
        cls_SHOW/1,
        proc_ReplaceCLS/4,
        pg_ReplacePROC/3,
        cls_READ/2,
        pg_AddCLS/3
     ]).

:- use_module(substitution,	[sub_display/1]).	
:- use_module(kb,		[valid_kb/1]).
:- use_module(aux,		[ask_oneof/3]).



/*corrector(+PgIn,+Bug,-PgOut) user interface for correcting bugs.*/

corrector(P,unknown,P).

corrector(Pgi,inconsistent(Ci),Pgo) :- 
    nl,
    pg_CLAUSETEMPLATE(Ci,Pgi,Pri,Cim,Cw,SUB,Rk),
    (
        \+ SUB = [] 
    ->
        write('This is an inconsistent clause.'),nl,
        cls_SHOW(Cim),
        write('It is an instance of '),nl,
        cls_SHOW(Cw),
        write('with '),nl,
        sub_display(SUB)
    ;
        write('This is an inconsistent clause.'),nl,
        cls_SHOW(Cim)
    ),
    ask_oneof('discard/modify/none? ', [discard,modify,none],  ACTION),
    (  
        ACTION == none 
    ->  
        Pgo = Pgi
    ;
        ACTION == modify
    -> 
        cls_MODIFY(Cw,Co),
        proc_ReplaceCLS(Co,Rk,Pri,Pro),
        pg_ReplacePROC(Pro,Pgi,Pgo)
    ;
        ACTION == discard 
    -> 
        proc_DeleteCLS(Rk,Pri,Pro),
        pg_ReplacePROC(Pro,Pgi,Pgo)
    ).
        
corrector(Pgi,uncovered(Atom),Pgo) :- 
    functor(Atom,Fn,Fa),
    (
        pg_PROC(Pgi,Fn/Fa,Pri)
    ->
        nl,tab(4),write(Atom),write(' is not completely covered by'),nl,nl, 
        proc_SHOW(Pri),
        ask_oneof('add/modify/none? ', [add,modify,none],  ACTION),
        (  
            ACTION == none
        ->
            Pgo = Pgi
        ;
            ACTION == add
        ->  
            proc_ReadCLS(Pri,Pro),
            pg_ReplacePROC(Pro,Pgi,Pgo)
        ;
            ACTION == modify
        ->
            proc_MODIFY(Pri,Pro),
            pg_ReplacePROC(Pro,Pgi,Pgo)
        )
    ;
        write(' Procedure '), write(Fn/Fa), write(' is not defined.'),nl,
        ask_oneof('add/none? ', 
                  [add,none],  ACTION),
        (  
            ACTION == none
        ->
            Pgo = Pgi
        ;
            ACTION == add
        ->  
            write('New Clause >>'),
            cls_READ(user_input,Cls),
            pg_AddCLS(Pgi,Cls,Pgo)
        )
    ).



corrector(Pgi,diverging_invocation((Pt0,St0),(Pt1,St1)),Pgo) :- 
    nl,
    write('This is a diverging invocation.'),nl,
    pg_ShowSTATE(Pgi,Pt0,St0),
    nl,
    pg_ShowSTATE(Pgi,Pt1,St1),
    ask_oneof('first/second? ', [first,second],ANSWER),
    (
        ANSWER = first
    ->
        Pt0 = (((PN/PA),RK),_)
    ;
        Pt1 = (((PN/PA),RK),_)
    ),    
    ask_oneof('discard/modify/none? ', [discard,modify,none], ACTION),
    pg_PROC(Pgi,PN/PA,Pri),
    proc_CLS(Pri,RK,Cd),
    (  
        ACTION == none
    ->  
        Pgo = Pgi
    ;
        ACTION == modify
    -> 
        cls_MODIFY(Cd,Co),
        proc_ReplaceCLS(Co,RK,Pri,Pro),
        pg_ReplacePROC(Pro,Pgi,Pgo)
    ;
        ACTION == discard 
    -> 
        proc_DeleteCLS(RK,Pri,Pro),
        pg_ReplacePROC(Pro,Pgi,Pgo)
    ).
        



