:- module(kb,[
    judge/3,
    valid_kb/1,
    faithful/1,
    complete/1,
    correct/1,
    store/2,
    bip/1,
    clear/0]).

:- use_module(meta,[solve_db/3]).

% The following is a list of the Built-In-Predicates 

bip('display'(_TERM)).      % built-in predicates for input/output
bip('nl').                  % they do not change program state
bip('print'(_TERM)).
bip('read'(_TERM)).         % except read that changes program state
bip('tab'(_Natural)).
bip('write'(_TERM)).
bip('writeq'(_TERM)).

bip('atom'(_TERM)).         % built-in predicates for test
bip('atomic'(_TERM)).       % they do not change program state
bip('integer'(_TERM)).
bip('number'(_TERM)).
bip('novar'(_TERM)).
bip('var'(_TERM)).

bip('=:='(_NUMBER1,_NUMBER2)).
bip('=\='(_NUMBER1,_NUMBER2)).
bip('<'(_NUMBER1,_NUMBER2)).
bip('>'(_NUMBER1,_NUMBER2)).
bip('=<'(_NUMBER1,_NUMBER2)).
bip('>='(_NUMBER1,_NUMBER2)).
bip('@<'(_TERM1,_TERM2)).
bip('@>'(_TERM1,_TERM2)).
bip('@=<'(_TERM1,_TERM2)).
bip('@>='(_TERM1,_TERM2)).

bip('='(_Left,_Right)).
bip('=='(_Left,_Right)).
bip('\=='(_Left,_Right)).


bip('true').                  % flow-control
bip('false').
bip('fail').

bip('is'(_TERM,_NUMBER)).     % may change program state


% The following predicates are related to assertions

:- dynamic '$faithful'/1.
:- dynamic '$correct'/1.
:- dynamic '$valid'/1.
:- dynamic '$invalid'/1.
:- dynamic '$complete'/1.


complete(A) :-
    \+ \+ (numbervars(A,0,_),'$complete'(A)).

faithful(A) :- 
    \+ \+ (numbervars(A,0,_),bip(A)).
faithful(A) :-
    \+ \+ (numbervars(A,0,_),'$faithful'(A)).

correct(A) :-
    faithful(A).
correct(A) :-
    \+ \+ (numbervars(A,0,_),'$correct'(A)).

% The following predicates are related to partial intended interpretation

valid_kb(A) :-
    '$valid'(A).

store(valid,A)   :-
    assertz('$valid'(A)).
store(complete,A) :-
    assertz('$complete'(A)).
store(invalid,A)  :-
    numbervars(A,0,_),
    assertz('$invalid'(A)).

store(unsatisfiable,A) :-  % needs to check consistency with I'.
    assertz('$complete'(A)).

store(faithful,A) :-  
    assertz('$faithful'(A)).
store(correct,A) :-  
    assertz('$correct'(A)).



clear :-
    ( retract('$valid'(_)),fail; true ),
    ( retract('$invalid'(_)),fail; true ),
    ( retract('$complete'(_)),fail; true ),
    ( retract('$faithful'(_)),fail; true ),
    ( retract('$correct'(_)),fail; true ).

judge(forall,\+A,P) :-
    !,
    judge(notexists,A,P).
judge(forall,A,P) :- 
        /*\forall A <= faithful(A) and \forall A is provable from P.*/
    \+ A = \+(_),
    faithful(A),
    \+ \+ (numbervars(A,0,_),solve_db(P,A,true)).
judge(forall,A,_) :- 
        /*A is an instance of a valid atom, so A is valid.*/
    \+ A = \+(_),
    \+ \+ (numbervars(A,0,_),'$valid'(A)).

judge(notall,A,_) :-
        /*A has an instance that is invalid.*/
    \+ \+ '$invalid'(A).
judge(notall,A,_) :-
        /*A is an instance of an atom of which all the valid 
           instances are known and A is not an instance of 
           any of those valid instances.*/        
    \+ \+ (numbervars(A,0,_),'$complete'(A)),
    \+ (numbervars(A,0,_),'$valid'(A)).
judge(notall,A,P) :-
        /*\not\forall A <= faithful(A) and P finitely fails on \forall A.*/
    faithful(A),
    \+ (numbervars(A,0,_),solve_db(P,A,_)).

judge(notexists,(L,W),P) :-
    (
        judge(notexists,L,P)
    ;
        judge(notexists,W,P)
    ).
judge(notexists,\+A,P) :-	% \not\exists\not A <==> \forall A
    judge(forall,A,P).
judge(notexists,A,P) :-	
    /*\not\exists A <= faithful(A) and P finitely fails on A.*/
    \+ A = \+(_),
    faithful(A),
    \+ solve_db(P,A,_). % P finitely fails on A
judge(notexists,A,_) :- 
    /*The set of all valid instances of A is empty.*/
    \+ A = \+(_),
    \+ \+ (numbervars(A,0,_),'$complete'(A)),
    \+ '$valid'(A).   


