:- module(meta,[solve_db/3]).

:- use_module(kb,[bip/1]).

:- use_module(parameters,[read_parameter_value/2]).

:- use_module(substitution,[
        variant/2]).

:- use_module(aux,[term_rename/2,vars/2]).

:- use_module(store,[pg_MatchCLS/3,pg_EnumerateCLS/4]).


/* 

solve_db(+P,?A,-S)  is a depth bound Prolog meta-interpreter.
solve_db(P,A,S)  solves  +A  with +P  using  depth-first  and
left-to-right  strategy  within  a  depth  bound.
solve_db(P,A,S)  succeeds  with  -A instantiated to  a solved
instance of +A and -S to "true" until a run-time error is detected. If
a  run-time error is detected,  solve_db(P,A,S) succeeds with
-S being err(Cat)  where Cat indicates the  category  of the
error.

*/

solve_db(P,A,S) :-
    read_parameter_value(depth_bound,D),
    solve_db(P,A,D,S),
    (
       S \== true,
       !  % prevent backtracking from taking place.
    ;
       true
    ).

/*

solve_db(+P,?A,+D,-S) is a depth bound meta-interpreter for normal
logic programs.  solve_db(P,A,D,S) solves goal +A with program +P
using depth-first and left-to-right strategy within depth bound D.
solve_db(P,A,D,S) succeeds if there is a SLDNF refutation for
$P\cup\{\leftarrow A\}$ within depth bound D with ?A being
instantiated to a solved instance of ?A and -S to "true".
solve_db(P,A,D,S) also succeeds if a run-time error is detected with
-A being partially instantiated and S being err(Cat).

*/

solve_db(_P,_A,0,err(overflow)) :- 
    !.

solve_db(P,\+(A),D,S) :-
    D>0,
    !,
    D_ is D-1,
    term_rename(A,A1),                   
    (   
        solve_db(P,A1,D_,Sa)
    ->
        (   
            Sa = true
        -> 
            (
                \+ variant(A,A1) 
            ->
                S = err(unsafe)
            )
        ; 
            S = Sa
        )
    ;   
        S=true  
    ).


solve_db(P,(A,B),D,S)          :- 
    D>0,
    !,
    solve_db(P,A,D,Sa),
    (
       Sa = true
    -> 
       solve_db(P,B,D,S)
    ; 
       S = Sa
    ).

solve_db(_P,A,_D,true) :-     % predicate true is a bip
    bip(A),
    !,
    call(A).

solve_db(P,A,D,S)             :-
    D_ is D - 1,
    pg_MatchCLS(A,P,B), 
    solve_db(P,B,D_,S).

