:- module(oracle, [
                   admissible/3,
                   valid/2,
                   satisfiable/2,
                   query_exists/2,
                   examine_solutions/4]).


:- use_module(store,[pg_ShowSTATE/3]).

:- use_module(aux,[bagof0/3,nodup/2,term_rename/2,vars/2,writel/1,yesno/1]).

%:- ensure_loaded(library(prompt)).
%:- ensure_loaded(library(ask)).

:- use_module(substitution,[
       sub_constructor/3,
       subs_display/1,
       sub_genSymbTb/3,
	sub_skolemize/3,
       sub_apply/3,
       subs_restrict/3,
       variant/2]).

:- use_module(parameters,[read_parameter_value/2]).

:- use_module(kb,[bip/1,faithful/1,complete/1,judge/3, store/2, valid_kb/1]).

/*valid(+L,+P) <==> \forall L is true in the intended interpretation
of P*/

valid(\+A,P) :-   % \forall\not A <==> \not\exists A.
    !,
    \+ satisfiable(A,P).
valid(A,P) :-	% A is known to be valid.
%    \+ A = \+(_),
    judge(forall,A,P).		
valid(A,P) :-	% A  is known to be invalid.
%    \+ A = \+(_),
     judge(notall,A,P), 		
     !,
     fail. 
valid(A,_) :- 	% The validity of A is not known.
%    \+ A = \+(_),
     query_forall(A,true).      


/*

query_forall(+A,-TruthValue) queries the user about the validity  of +A. If +A
is valid in the intended interpretation according to the user then -TruthValue
is set to "true". Otherwise -TruthValue is set to "false".

*/

query_forall(A,TruthValue) :- 
    sub_skolemize(A,[],A0),
    (   
	yesno(['Is ',A0,' valid'])
    -> 
        store(valid,A),
        TruthValue=true
     ; 
        store(invalid,A),
        TruthValue=false
    ).

/*satisfiable(?W,P) is  true iff  +W is  satisfiable in  the  intended
interpretation of P with -W is  instantiated into a valid  instance of
+W.*/

satisfiable((L,W),P) :-     % conjuction of literals 
    !,
    satisfiable(L,P),
    satisfiable(W,P).
satisfiable(\+A,P)   :-     % negative literal
    !,                         
    \+ valid(A,P).
satisfiable(A,P) :-      
%    \+ A = (_,_),
%    \+ A = \+(_),
    ground(A),
    !,
    valid(A,P).
satisfiable(A,P) :-
%    \+ A = (_,_),
%    \+ A = \+(_),
%    \+ ground(A),
    ( 
        check_exists(A,P,TruthValue),
        ( 
            TruthValue = true 
        -> 
            true
        ;
            !, %red cut           % all valid instances of A are aready known
            fail
        )
    ;
        query_exists(A,true)
    ). 

/* check_exists(?A,+P,-TruthValue) succeeds with -A being instantiated
into each valid instance of +A that is in the acquired part of the
intended interpretaion of P, and -TruthValue to "true". If all the
valid instances of +A are in the acquired part of the intended
interpretaion then check_exists(A,TruthValue) finally succeeds with -A
being +A and -TruthValue being "false". Otherwise, it fails straightly
after enumerating all the valid instances of +A that are in the
acqured part of the intended interpretation.

*/

check_exists(A,P,TruthValue) :-      
    faithful(A),               
    !,                               
    ( 
        solve_db(P,A,R)
    ->
        R = true,           % P succeeds on A
        TruthValue = true
    ;
        TruthValue = false  % P finitely fails
    ).
check_exists(A,_P,TruthValue)       :-
%    \+ faithful(A),
    (
        valid_kb(A), % A has a valid instance.
        TruthValue = true
    ;   
        complete(A),% A does not have any unknown valid instance
        TruthValue = false
    ).

/*

query_exists(?A,-TruthValue) queries the user  if +A has a valid instance that
is not in the acquired part of the intended  interpretation.  If there is such
an  instance then -TruthValue is set to "true"  and the user  is  requested to
provide   such  an  instance.  Otherwise  -TruthValue   is  set  to   "false".
Backtracking  to  query_exists(A,TruthValue)  after  a  "true"  is  given   to
-TruthValue results in the user being repeatedly queried while Backtracking to
query_exists(A,TruthValue) after  a "false" is given to  -TruthValue causes it
to fail.

*/

query_exists(A,TruthValue) :- 
    (
        complete(A) 
    ->  
        TruthValue = false
    ;
        (
            query_more(A,TruthValue) 
        ;
            query_exists(A,TruthValue)
        )
    ).



query_more(A,TruthValue) :-
    ground(A),
    !,
    (
        ( \+ valid_kb(A), yesno(['Is ',A,' valid']) )
    -> 
        store(valid,A),
        TruthValue = true
    ;  
        store(complete,A),
        TruthValue = false
    ).

query_more(A,TruthValue) :-
%    \+ ground(A),
    vars(A,VL),        % collect variables in A 
    sub_genSymbTb(VL,[],MSub), % generate mnemoric names for variables in A
    sub_apply(MSub,VL,VL1),  % replace  variables with mnemoric names
    sub_apply(MSub,A,A1),    % instantiate A with the mnemoric names 
    sub_constructor(VL1,VL,Sub),   
    bagof0(Sub,kb:valid_kb(A),Intended_Solution_List_So_Far),
    length(Intended_Solution_List_So_Far,Size),
    (
        Size == 0
    ->
        (   
            yesno(['Is ',A1,' satisfiable'])
        -> 
            get_var_val(MSub),
            store(valid,A),
            TruthValue = true
        ;  
            store(complete,A),
            TruthValue = false
        )
    ;
        (
            Size = 1
        ->
            writel(['This is a valid solution for ', A1, '.',nl])
        ;
            writel(['These are valid solutions for ', A1, ' .',nl])
        ),
        subs_display(Intended_Solution_List_So_Far),
        (
            yesno(['Is there any other solution for ',A1])
        ->
            get_var_val(MSub),
            store(valid,A),
            TruthValue = true
        ;
            store(complete,A),
            TruthValue = false
        )   
    ).
         

examine_solutions(A,S,P,Result) :-
    vars(A,VL),        % collect variables in A 
    subs_restrict(S,VL,S0),
    nodup(S0,S1),
    sub_genSymbTb(VL,[],MSub),  % generate mnemoric names for variables in A
    ( 
        wrong_solution(A,S1,P,Aw) 
    ->
        Result = declarative(wrong,Aw)
    ;
        missing_solution(A,S1,Am)
    ->
        Result = declarative(miss,Am)
    ;
        complete(A) 
    -> 
        Result = complete
    ;   
        ground(A)
    ->
        (
            valid(A,P) 
        ->
            Result = incomplete
        ;
            Result = complete
        )
    ;      
        sub_apply(MSub,A,A1), % instantiate A with the mnemoric names 
        sub_apply(MSub,S1,S2),
        length(S2,Size),
        (  
            Size == 0
        ->
            (
                yesno(['Is ',A1,' satisfiable'])
            -> 
                (
                   read_parameter_value(query_instance,yes)
                ->
                    get_var_val(MSub),
                    store(valid,A),
                    Result = declarative(miss,A)
                ;
                    Result=incomplete
                )
            ;
                Result=complete,
                store(complete,A)
            ) 
        ;
            ( 
                Size = 1
            ->
                writel(['This is a valid solution for ', A1, '.',nl]) 
            ;
                writel(['These are valid solutions for ', A1, '.',nl]) 
            ),
            subs_display(S2),
            (
                yesno(['Is there any other solution for ', A1])
            -> 
                (
                    read_parameter_value(query_instance,yes)
                ->
                    get_var_val(MSub),
                    store(valid,A),
                    Result = declarative(miss,A)
                ;
                    Result=incomplete
                )
            ;
                Result=complete,
                store(complete,A)
            ) 
        )
    ).

/*

wrong_solution(+Atom,+Subs,P,-InvalidInstance) +Atom is an atom,
+Sub_List is a list of substitutions whose domains are the set of
variables in +Atom.  wrong_solution(Atom,Subs,P,InvalidInstance)
succeeds iff there is a subsitution $\theta$ in +Subs such that
$Atom\theta$ is invalid in the intended interpretation with
$Invalid_Atom=Atom\theta$

*/

wrong_solution(A,[Sub|Sub_List],P,A1) :-
    sub_apply(Sub,A,A1),
    ( 
        \+ valid(A1,P) 
    ->  
        true
    ;
        wrong_solution(A,Sub_List,P,A1)
    ).
    
missing_solution(A,Sub_List,A1) :-
    all_instances(A,Sub_List,Instance_List),
    term_rename(A,A1),
    valid_kb(A1),
    \+ ( numbervars(A1,0,_),member(A2,Instance_List), A1=A2).

all_instances(_,[],[]).
all_instances(A,[S|SL],[I|IL]) :-
    sub_apply(S,A,I),
    all_instances(A,SL,IL).
    

%PRIVATE get_var_val/1

get_var_val([]).
get_var_val([Var/Name|Sub]) :-
    get_val(Name,Var),
    get_var_val(Sub).

get_val(VN,V) :-
        writel(['    ',VN,' = ? ']),
        read(V1),
        (
            V1 \== end_of_file 
        ->
            V = V1
        ;
            nl,
            get_val(VN,V)
        ).
        
admissible(P,Pt,Sub) :- 
    write('Look at the following program state.'),nl,
    pg_ShowSTATE(P,Pt,Sub),
    yesno(['Is that state admissible']).

