:- module(ordsets, [ 
	ord_setproduct/3,
	ord_subset/2,
	ord_add_element/3,
	ord_intersection/3,
	ord_subtract/3,
	ord_union/3
   ]).

ord_union([],S,S) :-
    !.
ord_union(S,[],S) :-
    !.
ord_union([H1|S1],[H2|S2],[H1|S3]) :-
    H1 @< H2,
    !,
    ord_union(S1,[H2|S2],S3).
ord_union([H1|S1],[H2|S2],[H1|S3]) :-
    H1 == H2,
    !,
    ord_union(S1,S2,S3).
ord_union([H1|S1],[H2|S2],[H2|S3]) :-
    H1 @> H2,
    !,
    ord_union([H1|S1],S2,S3).

ord_add_element([],E,[E]).
ord_add_element([E1|S1],E,[E,E1|S1]) :-
    E @< E1,
    !.
ord_add_element([E1|S1],E,[E1|S1]) :-
    E == E1,
    !.
ord_add_element([E1|S1],E,[E1|S2]) :-
    ord_add_element(S1,E,S2).

ord_setproduct([],_,[]).
ord_setproduct([E|S1],S2,P) :-
    scaling(E,S2,P1),
    ord_setproduct(S1,S2,P2),
    ord_union(P1,P2,P).

scaling(_,[],[]).
scaling(E,[E1|S1],[E-E1|S2]) :-
    scaling(E,S1,S2).

ord_subset([],_S) :-
    !.
ord_subset([H1|S1],[H2|S2]) :-
    H1 == H2,
    !,
    ord_subset(S1,S2).
ord_subset([H1|S1],[H2|S2]) :-
    H1 @> H2,
    !,
    ord_subset([H1|S1],S2).

ord_intersection([],_S,[]) :-
    !.
ord_intersection(_S,[],[]) :-
    !.
ord_intersection([H1|S1],[H2|S2],S3) :-
    H1 @< H2,
    !,
    ord_intersection(S1,[H2|S2],S3).
ord_intersection([H1|S1],[H2|S2],[H1|S3]) :-
    H1 == H2,
    !,
    ord_intersection(S1,S2,S3).
ord_intersection([H1|S1],[H2|S2],S3) :-
    H1 @> H2,
    !,
    ord_intersection([H1|S1],S2,S3).

ord_subtract([],_S,[]) :-
    !.
ord_subtract(S,[],S) :-
    !.
ord_subtract([H1|S1],[H2|S2],[H1|S3]) :-
    H1 @< H2,
    !,
    ord_subtract(S1,[H2|S2],S3).
ord_subtract([H1|S1],[H2|S2],S3) :-
    H1 == H2,
    !,
    ord_subtract(S1,S2,S3).
ord_subtract([H1|S1],[H2|S2],S3) :-
    H1 @> H2,
    !,
    ord_subtract([H1|S1],S2,S3).

