:- module(parameters,
       [read_parameter_value/2,set_parameter/0,view_parameters/0]).

:- use_module(aux,[ask_oneof/3,bagof0/3,get_term/2,writel/1]).

%:- ensure_loaded(library(prompt)).
%:- ensure_loaded(library(ask)).

:- dynamic '$para'/3.

% set default values for system parameters

'$para'(depth_bound,     20, lambda(X,(integer(X),X>=0,X=<20))).
'$para'(rank_bound,      10, lambda(X,(integer(X),X>=0,X=<10))).
'$para'(complexity_bound, 10, lambda(X,(integer(X),X>=0,X=<10))). 
'$para'(complexity_func, weight, lambda(X,member(X,[weight,height]))). 
'$para'(query_instance, no, lambda(X,member(X,[yes,no]))). 
'$para'(wrong_search_strategy, top_down_zooming, 
    lambda(X,member(X,[bottom_up,divide_and_conquer,top_down,top_down_zooming]))). 
'$para'(miss_search_strategy, top_down_zooming, 
    lambda(X,member(X,[bottom_up,divide_and_conquer,top_down,top_down_zooming]))). 

/*

read_parameter_value(+Para_Name,-Para_Value) 

*/

read_parameter_value(Para_Name,Para_Value) :-
    (
       '$para'(Para_Name,Para_Value_1,_)
    -> 
       Para_Value = Para_Value_1
    ;
       writel([Para_Name,' is not a system parameter',nl])
    ).

/*

set_parameter(+Para_Name,+Para_Value) 

*/

set_parameter(Para_Name,Para_Value)  :-
    ( 
        '$para'(Para_Name,_,C)
    ->
        (
            verify_condition(Para_Value,C)
        ->
            retract('$para'(Para_Name,_,_)),
            assertz('$para'(Para_Name,Para_Value,C))
        ;   
            writel([Para_Value,' is not valid for ', Para_Name,'!',nl])
        )
    ;  
        writel([Para_Name,' is not a system parameter',nl]) 
    ).

/*

view_parameters 

*/

view_parameters :- 
    nl,
    tab(8),write('***********************************************************'),nl,
    tab(8),write('*           CURRENT VALUES OF SYSTEM PARAMETERS           *'),nl,
    tab(8),write('*---------------------------------------------------------*'),nl,
    bagof0(Para_Name,parameters:Value^Cond^'$para'(Para_Name,Value,Cond), 
           Para_Name_List),
    sort(Para_Name_List,Sorted_Para_Name_List),
    view_parameters_one_by_one(Sorted_Para_Name_List),
    tab(8),write('***********************************************************'),nl.

/*

set_parameter

*/

set_parameter :- 
    view_parameters,nl,
    bagof0(Para,parameters:Value^Cond^'$para'(Para,Value,Cond), Para_List),
    ask_oneof('Which parameter? ', Para_List, Para_Name),
    get_term(Para_Value,['What value for ', Para_Name,' ? ']),
    set_parameter(Para_Name,Para_Value).

verify_condition(Value,lambda(Var,Cond)) :-
    \+ \+ (Var=Value,call(Cond)).

view_parameters_one_by_one([]).
view_parameters_one_by_one([Para_Name|Para_Name_List]) :-
       '$para'(Para_Name,Para_Value,_),
       tab(8),write('*'),tab(4),
           write(Para_Name),
           name(Para_Name,Para_Name_String),
           length(Para_Name_String,Width1),
           Width1_ is 24-Width1,
           tab(Width1_),
       write('='),tab(4),
           write(Para_Value),
           length_of_ground_term(Para_Value,Width2),
           Width2_ is 24-Width2,
           tab(Width2_),write('*'),
       nl,
       view_parameters_one_by_one(Para_Name_List).

length_of_ground_term(T,L) :-
    atomic(T),
    !,
    name(T,TN),
    length(TN,L).
length_of_ground_term(T,L) :-
%    \+ atomic(T),
    T =.. [F|ArgList],
    name(F,FN),
    length(FN,L0),
    length_of_ground_term_list(ArgList,L1),
    L is L0+2+L1. % 2 for ()
length_of_ground_term_list([],0).
length_of_ground_term_list([T],L) :-
    length_of_ground_term(T,L).
length_of_ground_term_list([T1,T2|TL],L) :-
    length_of_ground_term(T1,L1),
    length_of_ground_term_list([T2|TL],L2),
    L is L1+1+L2. % 1 for comma
       
