:- module(store,[
	            pg_LOAD/2,
        	    pg_SAVE/2,
	            pg_FUNCS/2,
	            pg_ShowSTATE/3,
	            pg_ShowINVOCATION/5,
	            pg_EMPTY/1,
        	    pg_PIDS/2,
            	    pg_AddCLS/3,
	            pg_SHOW/1,
	            pg_EnumerateCLS/4,
	            pg_ShowPROCS/2,
	            pg_ShowPROC/2,
        	    pg_PROC/3,
	            pg_AddPROC/3,
	            pg_MatchCLS/3,
	            pg_ReplacePROC/3,
	            pg_CLAUSETEMPLATE/7,
	            proc_SHOW/1,
	            proc_ReadCLS/2,
	            proc_MODIFY/2,
	            proc_ReplaceCLS/4,
	            proc_DeleteCLS/3,
	            proc_SIZE/2,
	            proc_PID/2,
	            proc_CLS/3,
	            cls_SIZE/2,
	            cls_ANNOTATION/2,
	            cls_IMPLICATION/2,
	            cls_SYMBOLTABLE/2,
	            cls_MODIFY/2,
	            cls_SHOW/1,
        	    cls_CONS/4,
	            cls_READ/2
         ]).

:- use_module(goal,[goal_compose/3]).

:- use_module(substitution, [
                 sub_genSymbTb/3,
                 sub_skolemize/3,
                 sub_apply/3,
                 sub_inverse/2,
                 mgu/3,
                 sub_domain/2
             ]).

:- use_module(aux,[bagof0/3,term_rename/2,vars/2,ask_number/4]).

:- use_module(ordsets,	[
	ord_add_element/3,
	ord_subtract/3,
	ord_union/3
   ]).



%:- ensure_loaded(library(prompt)).
%:- ensure_loaded(library(ask)).



%%%%%%\\\\\///// local procedures \\\\\/////%%%%%%

/*locate_indexed_cl(+Lst,+Rank,-Head,-Tail)  append(Head,Tail,Lst) and
-Head  consists of the  first (Rank-1)  elements of +Lst  if  +Rank is
appropriately given. */

locate_indexed_cl([],_,[],[]).
locate_indexed_cl([Ele|Lst],Rank,[],[Ele|Lst]) :-
        Rank =< 1.
locate_indexed_cl([Ele|Lst],Rank,[Ele|Head],Tail) :-
        Rank>1,
        Rank1 is Rank-1,
        locate_indexed_cl(Lst,Rank1,Head,Tail).


/*nameListToSub(+VL,+NL,-SUB) -SUB  is  a substitution over  domain VL
with variables assigend  their names as defined  in NL or an anonymous
names if undefined in NL.*/

nameListToSub(VL,NL,Sub) :-
    nameListToSub_1(NL,Sub1),
    sub_domain(Sub1,VL1),
    ord_subtract(VL,VL1,VL2),
    anonymous_sub(VL2,Sub2),
    ord_union(Sub1,Sub2,Sub). 

nameListToSub_1([],[]).
nameListToSub_1([Name=Var|NL1],Sub) :-
    nameListToSub_1(NL1,Sub1),
    ord_add_element(Sub1,Var/Name,Sub).

anonymous_sub([],[]).
anonymous_sub([Var|Vars],[Var/'_'|Sub]) :-
    anonymous_sub(Vars,Sub).

/*pg_PIDS(+Pg,-DescriptorsOfDefinedPredicates) */

pg_PIDS(prog(ProcLst),DesciptorLst) :-
    pg_PIDS_1(ProcLst,DesciptorLst).
pg_PIDS_1([],[]).
pg_PIDS_1([proc(PN,PA,_)|ProcLst],[PN/PA|DesciptorLst]) :-
    pg_PIDS_1(ProcLst,DesciptorLst).
    

/* pg_CLAUSETEMPLATE(+(H:-B),+Pg,-Pr,-Cmnem,-Cl,-SUB-Rk)*/

/*
pg_CLAUSETEMPLATE((H:-B),Pg,proc(PN,PA,CLS),Cmnem,Cl,SUB,Rk) :-
    sub_genSymbTb((H:-B),[],ALPHA),
    sub_apply(ALPHA,(H:-B),HBm),
    cls_CONS(HBm,[],[],Cmnem),
    functor(H,PN,PA),
    pg_PROC(Pg,PN/PA,proc(PN,PA,CLS)),
    pg_CLAUSETEMPLATE_1(HBm,CLS,Cl,SUB,1,Rk).

pg_CLAUSETEMPLATE_1(HB,[Cl1|CLS],Cl,SUB,Rk0,Rk) :-
    cls_IMPLICATION(Cl1,HB1),
    (
        mgu(HB,HB1,ALPHA)
    ->
        cls_SYMBOLTABLE(Cl1,SUB1),
        Cl = Cl1,
        sub_apply(SUB1,ALPHA,SUB),
        Rk = Rk0
    ;
        Rk1 is Rk0 + 1,
        pg_CLAUSETEMPLATE_1(HB,CLS,Cl,SUB,Rk1,Rk)
    ). 
*/


pg_CLAUSETEMPLATE((H:-B),P,proc(PN,PA,Cs),Cm,Co,Sub,Rk) :-
    sub_skolemize((H:-B),[],HB0),
    functor(H,PN,PA),
    pg_PROC(P,PN/PA,proc(PN,PA,Cs)),
    find_cls(HB0,Cs,Co,1,Rk),
    cls_SYMBOLTABLE(Co,STB),
    sub_inverse(STB,InversedSTB),
    sub_domain(InversedSTB,Reserved),
    sub_skolemize((H:-B),Reserved,HBm),
    cls_CONS(HBm,[],[],Cm),
    cls_IMPLICATION(Co,HB),
    mgu(HBm,HB,ALPHA),
    sub_apply(STB,ALPHA,Sub).

find_cls(HB0,[C1|Cs],C,Rk0,Rk) :-
    cls_IMPLICATION(C1,HB1),
    (
        mgu(HB0,HB1,_)
    ->
        C = C1,
        Rk = Rk0
    ;
        Rk1 is Rk0 + 1,
        find_cls(HB0,Cs,C,Rk1,Rk)
    ). 


% predicates for object "pred"

proc_SIZE(proc(_,_,CLS),Size) :-
    length(CLS,Size).

proc_PID(proc(PN,PA,_),PN/PA).

proc_CLS(proc(_,_,CLS),Rk,Cl) :-
    proc_CLS(CLS,1,Rk,Cl).

proc_CLS([Cl|_],Rk,Rk,Cl). 
proc_CLS([_|CLS],N,Rk,Cl) :-
    N < Rk,
    N1 is N+1,
    proc_CLS(CLS,N1,Rk,Cl).

% predicates for object "Cl"

cls_IMPLICATION('$cl'(IMPL,_,_),IMPL).
cls_SYMBOLTABLE('$cl'(_,SYMBTAB,_),SYMBTAB).
cls_ANNOTATION('$cl'(_,_,ANNS),ANNS).
cls_CONS(IMPL,SYMBTAB,ANNS,'$cl'(IMPL,SYMBTAB,ANNS)).

/*cls_SHOW(+Cl) prints clause Cl in its mnemonic form*/
 
cls_SHOW(C) :-
    cls_IMPLICATION(C,HB),
    cls_SYMBOLTABLE(C,Sub),
    sub_apply(Sub,HB,HB1),
    impl_SHOW(HB1).

impl_SHOW((A:-B)) :-
    tab(4),write(A),
    ( 
        B==true 
    -> 
        write('.'), nl
    ;  
        write(' :-'),nl,
        body_SHOW(B)
    ).

body_SHOW((A,B)) :- 
    !, 
    tab(8),write(A),write(','),nl,
    body_SHOW(B).  
body_SHOW(A) :- 
    % \+ A=(_,_),
    tab(8),
    write(A),
    write('.'),nl.


cls_SIZE(Cl,Size) :-
    cls_IMPLICATION(Cl,(_:-B)),
    impl_size(B,Size).
impl_size(true,0).
impl_size(A,1) :-
    \+ A = true,
    \+ A =(_,_).
impl_size((A,B),Size) :-
    impl_size(A,Sa),
    impl_size(B,Sb),
    Size is Sa+Sb.

/*cls_READ(+Stream,-Clause)  reads a  clause -Clause  from an input
stream +Stream. If the end of the stream is encountered, Clause is set
to end_of_file*/

cls_READ(Stream,Clause) :-
    read_term(Stream,Term,[syntax_errors(dec10),variable_names(NL)]),
    (
        Term = end_of_file
    -> 
        Clause = end_of_file
    ;
	Term = (:- Call)
    ->
	call(Call),
        cls_READ(Stream,Clause)
    ;
        (
            Term = (H:-B1) 
        -> 
            true
        ;
            H = Term,
            B1 = true
        ),
        vars(Term,VL),
        nameListToSub(VL,NL,Sub),
        Clause = '$cl'((H:-B1),Sub,[])
     ).

cl_getPD(Cl,PN/PA) :-
    cls_IMPLICATION(Cl,(H:-_B)),
    functor(H,PN,PA).


read_pd(PDS,PD) :-
    write('Which predicate? '),
    read(PD1),
    (
        PD1 = (_/_)
    ->
        (
            member(PD1,PDS)
        ->
            PD=PD1
        ;
            write('%Predicate '),write(PD1),write(' is not defined.'),nl,
            read_pd(PDS,PD)
        )
    ;
        bagof0(PD1/PA,basics:member(PD1/PA,PDS),PDS1),
        length(PDS1,Size),
        (
            Size = 0
        ->
            write('%Predicate '),write(PD1),write(' is not defined.'),nl,
            read_pd(PDS,PD)
        ;
            Size = 1
        ->
            PDS1 = [PD]
        ;   
            write('%There are several procedure for '),write(PD1),write(' .'),nl,
            write('What arity? '),read(PA),
            (
                member(PD1/PA,PDS1)
            -> 
                PD = PD1/PA
            ;
                write('%Predicate '),write(PD1/PA),write(' is not defined.'),nl,
                read_pd(PDS,PD)
            )
        )
    ).

    


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pg_ShowSTATE(P,((Pn/Pa,RK),PtY),State) :-
    pg_PROC(P,Pn/Pa,Pred),
    proc_CLS(Pred,RK,Cl),
    cls_IMPLICATION(Cl,HB),
    cls_SYMBOLTABLE(Cl,SYMBTB),
    sub_apply(SYMBTB,HB,HB0),
    sub_apply(SYMBTB,State,State0),
    cls_gid_state_SHOW(HB0,PtY,State0).

pg_ShowINVOCATION(P,(PID,Rank),Entry,Gid,State) :-
    pg_PROC(P,PID,Pred),
    proc_CLS(Pred,Rank,Cl),
    cls_IMPLICATION(Cl,HB),
    cls_SYMBOLTABLE(Cl,SYMBTB),
    sub_apply(SYMBTB,HB,HB0),
    sub_apply(SYMBTB,Entry,Entry0),
    sub_apply(SYMBTB,State,State0),
    cls_state_gid_state_SHOW(HB0,Entry0,Gid,State0).

cls_state_gid_state_SHOW((H:-B),Entry,Gid,State) :- % Gid =\= 1
    tab(4),write(H),
    ( 
        B==true 
    -> 
        write('.'), nl,
        tab(12), state_SHOW(Entry), nl
    ;  
        write(' :-'),nl,
        tab(12), state_SHOW(Entry), nl,
        body_gid_state_SHOW(2,B,Gid,State)
    ).


body_gid_state_SHOW(Current,(A,B),Gid,State) :- 
    !, 
    tab(8),write(A),write(','),nl,
    ( Current = Gid -> tab(12), state_SHOW(State),nl ; true ),
    Next is Current+1,
    body_gid_state_SHOW(Next,B,Gid,State).  
body_gid_state_SHOW(Current,A,Gid,State) :- 
    % \+ A=(_,_),
    tab(8), write(A), write('.'),nl,
    ( Current=Gid -> tab(12), state_SHOW(State), nl ; true ).

cls_gid_state_SHOW((A:-B),K,Sub) :-
    tab(4),write(A),
    ( 
        B==true 
    -> 
        write('.'),nl,
        ( K=1 -> tab(12), state_SHOW(Sub), nl ; true )
    ;  
        write(' :-'),nl,
        ( K=1 -> tab(12), state_SHOW(Sub), nl ; true ),
        body_gid_state_SHOW(2,B,K,Sub)
    ).

state_SHOW([]).
state_SHOW([X/T|Sub]) :-
    write('  '),write(X),write(' <- '),write(T),
    state_SHOW(Sub).


/*A  prolog  program   is  represented   as  a   term   of  the   form
program(ListOfProcedures)  where  ListOfProcedures  is a list  of  the
representations of a set of procedres. A procedre is represented as  a
term  of  the  form proc(Name,Arity,ListOfClauses)  where  Name is the
name of  the  procedure,  Arity the  arity  of the  procedre  and  the
ListOfClauses is an ordered  list of the representation of the clauses
of the  procedure. A  clause  is represented  as  a  term of  the form
cl((Head:-Body),Sub)  where  (head:-Body)  is the  program  clause
in the form of internal representation of  the PROLOG system, that is,
the  variables are in  the  form of storage addresses,  where  Sub  is
maps the variable into their name */

pg_EMPTY(prog([])).

/*pg_LOAD(+File,-Program) reads -Program from +File */

pg_LOAD(File,Prog) :-
    clear,
    fileerrors(OldFileErrorsFlag,off), %turn off fileerrors
    ( 
        open(File,read,Stream)
    ->
        pg_LOAD_1(Stream,Prog),
        close(Stream)
    ;   
        nl,write(File),write(' does not exist!'),nl,
        pg_EMPTY(Prog)
    ),
    fileerrors(_,OldFileErrorsFlag).  %restore fileerrors

pg_LOAD_1(Stream,Prog) :-
    pg_LOAD_1(Stream,prog([]),Prog).
pg_LOAD_1(Stream,Pg0,Pg) :-
    cls_READ(Stream,Cl),
    (
        Cl = end_of_file
    ->
        Pg = Pg0       
    ;
        pg_AddCLS(Pg0,Cl,Pg1),
        pg_LOAD_1(Stream,Pg1,Pg)
    ).

pg_AddCLS(OldProg,Cl,NewProg) :-
    cl_getPD(Cl,PN/PA),
    (
        pg_PROC_pg(OldProg,PN/PA,OldProc,OldProg1)
    ->
        cls_proc_ADD_proc(Cl,OldProc,NewProc),
        pg_AddPROC(NewProc,OldProg1,NewProg)
    ;
        pg_AddPROC(proc(PN,PA,[Cl]),
                           OldProg,NewProg)
    ).

/*pg_AddPROC(+Proc,+ProgI,-ProgO) ProgO is ProgI plus Proc.*/

pg_AddPROC(Proc,prog(ProcList),prog([Proc|ProcList])).

/*cls_proc_ADD_proc(+Cl,+ProcI,-ProcO) procedure ProcO is procedure
ProcI plus clause Cl.*/

cls_proc_ADD_proc(Cl,proc(PN,PA,Cls0),proc(PN,PA,Cls1)) :-
    append(Cls0,[Cl],Cls1).

/*pg_PROC_pg(+ProgI,+PN/+PA,-Proc,ProgO) succeeds if ProgI
contains clauses for procedure PN/PA.  If so,  Proc is the  definition
of  PN/PA  and  ProgO  contains definitions  for other  procedures  in
ProgI.*/

pg_PROC_pg(prog([Proc|ProcList]),PN/PA,Proc,prog(ProcList)) :-
    Proc = proc(PN,PA,_),
    !.
pg_PROC_pg(prog([Proc0|ProcList0]),PN/PA,Proc,prog([Proc0|ProcList1])) :-
    pg_PROC_pg(prog(ProcList0),PN/PA,Proc,prog(ProcList1)).

/*pg_PROC(+PgI,+(PN/PA),-PrO)*/

pg_PROC(PgI,Pn/Pa,PrO) :- 
    pg_PROC_pg(PgI,Pn/Pa,PrO,_).

/*proc_SHOW(+Proc) dispalys clauses that define procedure Proc.*/

proc_SHOW(proc(PN,PA,Cls)) :-
    write('%PREDICATE --- '),
    write(PN/PA),
    write(' ---'),nl,nl,
    proc_SHOW_1(Cls),
    nl.

proc_SHOW_1([]).
proc_SHOW_1([Cl|Cls]) :-
        cls_SHOW(Cl),
        proc_SHOW_1(Cls).

/*pg_SAVE(+File,-Program) writes -Program into +File */

pg_SAVE(File,prog(PRS)) :-
    (    
        access_file(File,exist)
    ->  
        write('The file '),write(File), write(' already exists'),nl,nl
    ;
        open(File,write,Stream),
        set_output(Stream),
        pg_SAVE_1(PRS),
        close(Stream)
    ).

pg_SAVE_1([]).
pg_SAVE_1([PR|PRS]) :-
    proc_SHOW(PR),
    pg_SAVE_1(PRS).

%pg_pids_SHOW(+Pg,+PD_List).

pg_pids_SHOW(_,[]).
pg_pids_SHOW(Pg,[PD|PDL]) :-
    pg_ShowPROC(Pg,PD),
    nl,
    pg_pids_SHOW(Pg,PDL).

%list(+Program,+Pred/+Arity) lists clauses for Pred/Arity.

pg_ShowPROC(Program,Pn/Pa) :- 
    (
        pg_PROC(Program,Pn/Pa,Procedure)
    ->
        proc_SHOW(Procedure)
    ;
        write('%PREDICATE --- '),write(Pn/Pa),
        write(' ---- undefined.'),nl,nl
    ).

pg_ShowPROCS(Pg,PN) :-
    pg_PIDS(Pg,Defined),
    bagof0( PN/PA, basics:PA^member(PN/PA,Defined),Relevant),
    (
        Relevant = []
    -> 
        write('%Predicate --- '),write(PN),write(' --- undefined.'),nl,nl
    ;
        sort(Relevant,Relevant1),
        pg_pids_SHOW(Pg,Relevant1)
    ).

pg_SHOW(Pg) :- 
    pg_PIDS(Pg,Defined),
    (
        Defined = [] 
    ->
        write('%The program is empty.'),nl,nl
    ;
        sort(Defined,Defined1),  
        pg_pids_SHOW(Pg,Defined1)
    ).

/*pg_ReplacePROC(+Pred,+Pg,-PgN)*/

pg_ReplacePROC(Pred,prog(Prs),prog(PrsN)) :-
    pg_ReplacePROC_1(Pred,Prs,PrsN).

pg_ReplacePROC_1(proc(PN,PA,CL),[],[proc(PN,PA,CL)]).
pg_ReplacePROC_1(proc(PN,PA,CL),[proc(PN1,PA1,CL1)|Prs],PrsN) :-
    (
        PN=PN1,PA=PA1
    ->
        PrsN = [proc(PN,PA,CL)|Prs]
    ;
        pg_ReplacePROC_1(proc(PN,PA,CL),Prs,PrsN1),
        PrsN = [proc(PN1,PA1,CL1)|PrsN1]
    ).

/*proc_ReadCLS(+PrI,-PrO) PrO is PrI plus a clause from the
user at a user-specified position. */

proc_ReadCLS(proc(PN,PA,CLi),proc(PN,PA,CLo)) :- 
    write('Add clause: '),nl,
    cls_READ(user_input,C),
        (
            C = end_of_file 
        -> 
            CLo = CLi
        ; 
            ( 
                cl_getPD(C,PN/PA) 
            -> 
                length(CLi,Max),
                Max1 is Max+1,
                ask_number('Which clause', 1, Max1, Rank),
                locate_indexed_cl(CLi,Rank,CLi1,CLi2),
                append(CLi1,[C|CLi2],CLo)
            ;
                CLo = CLi,
                write('The input clause does not belong to procedure '),
                write(PN/PA),write('.'),nl
            )
        ).

/*proc_MODIFY(+PrI,-PrO)  A clause  in +PrI is modified -PrO
is the modified definition for the same procedure.*/

proc_MODIFY(proc(PN,PA,CLi),proc(PN,PA,CLo)) :-
        length(CLi,MaxRank),
        (  MaxRank == 1 ->
           Rank = MaxRank 
         ; ask_number('Which clause',1,MaxRank,Rank)
        ),
        locate_indexed_cl(CLi,Rank,CLi1,[Ci|CLi2]),
        cls_MODIFY(Ci,Co),
        append(CLi1,[Co|CLi2],CLo).        

proc_ReplaceCLS(Cl,Rk,proc(PN,PA,CLS),proc(PN,PA,CLSN)) :-
    length(CLS,MRk),
    (
        MRk >= Rk
    -> 
        proc_ReplaceCLS_1(Cl,Rk,CLS,CLSN)
    ;
        write('Illegal input to proc_ReplaceCLS/4.'),nl
    ).

proc_ReplaceCLS_1(C,1,[_|Cs],[C|Cs]).
proc_ReplaceCLS_1(C,Rk,[Ci|CsI],[Ci|CsO]) :-
    Rk>1,
    Rk1 is Rk-1,
    proc_ReplaceCLS_1(C,Rk1,CsI,CsO).


proc_DeleteCLS(Rk,proc(PN,PA,CLS),proc(PN,PA,CLSN)) :-
    length(CLS,MRk),
    (
        MRk >= Rk
    -> 
        proc_DeleteCLS_1(Rk,CLS,CLSN)
    ;
        write('Illegal input to proc_DeleteCLS/3.'),nl
    ).

proc_DeleteCLS_1(1,[_|Cs],Cs).
proc_DeleteCLS_1(Rk,[Ci|CsI],[Ci|CsO]) :-
    Rk>1,
    Rk1 is Rk-1,
    proc_DeleteCLS_1(Rk1,CsI,CsO).

        
/* pg_EnumerateCLS(PID,+P,-N,-(H:-B)) (H:-B) is a Nth
clause for A interspersed with invariants A */

pg_EnumerateCLS(PID,P,N,(H:-B)) :-
    pg_PROC(P,PID,proc(_,_,ListOfClause)),
    list_INDEXEDLIST(ListOfClause,ListOfIndexedClause),
    !,
    member((N,Clause),ListOfIndexedClause), % select a clause for PN/PA
    cls_IMPLICATION(Clause,(H:-B )). % find impl 
    
list_INDEXEDLIST(ListOfClause,ListOfIndexedClause) :-
    list_INDEXEDLIST(ListOfClause,1,ListOfIndexedClause).

list_INDEXEDLIST([],_N,[]).
list_INDEXEDLIST([Clause|ListOfClause],N,[(N,Clause)|ListOfIndexedClause]) :-
    N1 is N+1,
    list_INDEXEDLIST(ListOfClause,N1,ListOfIndexedClause). 

/*pg_MatchCLS(A,P,B) (A:-B) is an instance of a clause in P */

pg_MatchCLS(A,P,B) :-
    functor(A,PN,PA),
    pg_PROC(P,PN/PA,proc(PN,PA,CLS)), 
    !,
    member(Cl,CLS),             % select a clause for PN/PA
    term_rename(Cl,ClRenamed),         % rename the clause
    cls_IMPLICATION(ClRenamed,(H:-B)),
    A = H.                      % unify the head and the selected atom

/*cls_MODIFY(+Ci,-Co) modifies +Ci into -Co. */

cls_MODIFY(Ci,Co) :-
    write('Old clause: '),nl,
    cls_SHOW(Ci),
    write('New clause: '),nl,
    cls_READ(user_input,Co).        

pg_FUNCS(Pg,Fs) :-
    pg_PIDS(Pg,Pids),
    pids_pg_FUNCS(Pids,Pg,[],Fs).

pids_pg_FUNCS([],_,Fs,Fs).
pids_pg_FUNCS([Pid|Pids],Pg,Fs0,Fs1) :-
    pg_PROC(Pg,Pid,proc(_,_,Cs)),
    clss_FUNC(Cs,Fs0,Fs2),
    pids_pg_FUNCS(Pids,Pg,Fs2,Fs1).

clss_FUNC([],Fs,Fs).
clss_FUNC([C|Cs],Fs0,Fs1) :-
    cls_IMPLICATION(C,(H:-B)),
    literal_FUNCS(H,Fs0,Fs2),
    body_FUNCS(B,Fs2,Fs3),
    clss_FUNC(Cs,Fs3,Fs1).
    
literal_FUNCS(Literal,FunctorsAcc,Functors) :-
    (
        Literal = \+(Atom)
    -> 
        Atom =.. [_| Terms]
    ;
        Literal =.. [_| Terms]
    ),
    terms_FUNCS(Terms,FunctorsAcc,Functors).

body_FUNCS(true,FunctorsAcc,FunctorsAcc) :-
    !.
body_FUNCS((A,B),FunctorsAcc,Functors) :-
    !,
    literal_FUNCS(A,FunctorsAcc,NewFunctorsAcc),
    body_FUNCS(B,NewFunctorsAcc,Functors).
body_FUNCS(A,FunctorsAcc,Functors):-
    literal_FUNCS(A,FunctorsAcc,Functors). 

/*   terms_FUNCS(Terms,+FunctorsAcc,-Functors)   Functors   is
FunctorsAcc plus functors in Terms */

terms_FUNCS([],FunctorsAcc,FunctorsAcc).
terms_FUNCS([Term|Terms],FunctorsAcc,Functors)	:-
    term_FUNCS(Term,FunctorsAcc,NewFunctorsAcc),
    terms_FUNCS(Terms,NewFunctorsAcc,Functors).

/*    term_FUNCS(Term,+FunctorsAcc,-Functors)    Functors    =
FunctorsAcc plus functors in Term.*/
   
term_FUNCS(Term,FunctorsAcc,Functors)	:-
    (
        var(Term)
    -> 
        Functors = FunctorsAcc
    ;  
        functor(Term,Functor,Arity),
	Term =.. [Functor|Subterms],
        ord_add_element(FunctorsAcc,Functor/Arity,NewFunctorsAcc),
	terms_FUNCS(Subterms,NewFunctorsAcc,Functors)
    ).


