:- module(trace, [icppt/3,
                  icppt_RootImplication/2,
                  icppt_transformer/3,
                      icpst/3,
                      icpst_divisor/5,
                      icpst_transformer1/2,
                      icpst_transformer2/2,
                  aet/3]).

:- use_module(goal,[
        goal_compose/3,
        goal_select/3]).

:- use_module(orderedTree,[
        ordTree_constructor/4,
        forest_containingTree/5,
        forest_weight/2,
        forest_height/2]).



:- use_module(meta,[solve_db/3]).
:- use_module(aux,[bagof0/3,term_rename/2,vars/2]).
:- use_module(kb,[bip/1,judge/3,faithful/1,correct/1]).

:- use_module(substitution,[
        mgu/3,
        sub_apply/3,
        sub_compose/3,
        sub_restrict/3,
	variant/2
   ]).

:- use_module(parameters,	[
	read_parameter_value/2
   ]).

:- use_module(store,[pg_MatchCLS/3,pg_EnumerateCLS/4]).




/* 
SPECIFICATION
-------------

This  module  contains  procedures  that   construct  models  for  the
execution histories of bug symptoms of P wrt I.

PUBLIC icppt(+P,+A,-T) 
       icpst(+P,+A,-T)       
*/

/*icppt(+P,+A,-T) T is  an  I-congruent partial  proof  tree for P and
A.*/

icppt(P,L,T) :-
   term_rename(L,L1),                     % standarization apart
   icppt(L1,P,0,_,T).              % 0 labels the root of T 

/*icppt(+L,+P,+Id,-Idn,-T) T  is an I-congruent partial proof tree for
P and L.  Id  is the number for the root of T  and Idn is equal to one
plus the maxium number assigned to a node in T.*/

icppt(L,P,Id,Idn,t(Id,L,0,1,[])) :-
    judge(forall,L,P),
    !,
    Idn is Id + 1.

icppt(\+ A,P,Id,Idn,t(Id,\+A,1,2,[t(Id1,true,0,1,[])])) :- 
    !,
%    \+ judge(forall,\+A,P),
    Id1 is Id+1,
    Idn is Id+2,
    \+ solve_db(P,A,_). % P finitely fails on A.

icppt(A,P,Id,Idn,t(Id,A,H,W,CHDN)) :-  
%   \+ A = \+(_),  
%   \+ judge(forall,A,P),
   pg_MatchCLS(A,P,B),
   Id1 is Id+1,
   icppf(B,P,Id1,Idn,CHDN),
   forest_height(CHDN,Hmax), H is Hmax+1,
   forest_weight(CHDN,Wsum), W is Wsum+1. 

%icppf(+B,+P,+Id,-Idm,-T) 

icppf((L,B),P,Id,Idn,[T1|Ts]) :-
   icppt(L,P,Id,Idn_,T1),
   icppf(B,P,Idn_,Idn,Ts).
icppf(L,P,Id,Idn,[T]) :-
   \+ L=(_,_),
   icppt(L,P,Id,Idn,T).


/*icppt_transformer(+T1,+P,-T2)*/

icppt_transformer(t(Id,L,0,1,[]),_P,t(Id,L,0,1,[])) :-   % a leaf
    !.     
icppt_transformer(t(Id,L,_H,_,CHN),P,T2) :-
%    _H=\=0,
    (    
        judge(forall,L,P) 
    ->
        T2=t(Id,L,0,1,[])
    ; 
        icppf_transformer(CHN,P,CHN2),
        ordTree_constructor(Id,L,CHN2,T2)
    ).

icppf_transformer([],_P,[]).
icppf_transformer([T|Ts],P,[T1|Ts1]) :-
    icppt_transformer(T,P,T1),
    icppf_transformer(Ts,P,Ts1).

/*icppt_RootImplication(+T,-C): C is the root impl of T.*/

icppt_RootImplication(t(_,A,_,_,CHN),':-'(A,B)) :-
    icppf_RootConjunct(CHN,B).

icppf_RootConjunct([],true).
icppf_RootConjunct([t(_,A,_,_,_)],A) :-
    !.
icppf_RootConjunct([t(_,A,_,_,_)|Ts],(A,B)) :-
%   \+ Ts == [],
    icppf_RootConjunct(Ts,B).



/*icpstq(+P,+W,-T)   -T   is   an  $I$-complete  partial   SLDNF  tree
for P and $\leftarrow W$ via the left-to-right computation rule.*/

icpst(P,W,T) :-
    icpst(W/[],P,0,_,T).  % 0 labels the root of T

/*icpst(+W,+P,+Id,-Idn,-T)  -T is an  $I$-complete  partial SLDNF tree
for P and $\leftarrow  W$ via the left-to-right computation rule.  +Id
labels the root of T and -Idn is equal to one plus the maximun  number
that labels a node of T.*/

icpst(true/Sub,_,Id,Idn,t(Id,true/Sub,0,1,[])) :-
    !,
    Idn is Id+1.
icpst(G,P,Id,Idn,t(Id,G,Ht,Wt,Ts)) :-
        % G \== true,
    bagof0(G_,trace:P^derived(G,P,G_),Gs),
    Id_ is Id+1,
    icpsf(Gs,P,Id_,Idn,Ts),
    forest_height(Ts,Ht_),
    Ht is Ht_ + 1,
    forest_weight(Ts,Wt_), 
    Wt is Wt_ + 1.

icpsf([],_,Id,Id,[]).
icpsf([G|Gs],P,Id,Idn,F) :-
    (
        judge(notexists,G,P)
    ->
        icpsf(Gs,P,Idn_,Idn,F)   % prunes SLDNF tree for P and <- G
    ;
        icpst(G,P,Id,Idn_,T),
        icpsf(Gs,P,Idn_,Idn,Ts),
        F = [T|Ts]
    ).


/*derived(+Goal_In,+P,-Goal_Out)  -Goal_Out  is  a derived  goal  from
+Goal_In  and +P.  It is assumed  that, given an insufficiency symptom
$A$ of  $P$  {\sf wrt}  $I$,  no stack  overflow  will occur  when  an
$I$-complete   partial  SLDNF  tree  for  $P\cup\{\leftarrow  A\}$  is
constructed.   Should $A$ have caused a stack overflow, $A$  would not
be recognised as an insufficiency symptom of $P$ {\sf wrt} $I$.*/

derived(Goal_In/Sub_In,P,Goal_Out/Sub_Out) :-
    goal_select(Goal_In,L,Goal_In_1),
    (    L = \+(A) 
      -> 
         \+ solve_db(P,A,_),
         Goal_Out = Goal_In_1,
         Sub_Out  = Sub_In
       ;
         (    faithful(L) 
           -> 
              term_rename(L,L0),
              solve_db(P,L0,true),         % hyper-resolution step
              mgu(L,L0,Sub),
              sub_apply(Sub,Goal_In_1,Goal_Out),
              sub_compose(Sub_In,Sub,Sub_Out)
            ; 
              term_rename(L,H),
              pg_MatchCLS(H,P,B),   % resolution step
              mgu(L,H,Sub),
              goal_compose(B,Goal_In_1,Goal_Out_1),
              sub_apply(Sub,Goal_Out_1,Goal_Out),
              sub_compose(Sub_In,Sub,Sub_Out)
         )
    ).


/*

icpst_divisor(+CPS,+Id,-CPSa,-Sa,-CPS1) +CPS  is an I-complete  partial
SLDNF tree for $P\cup\{\leftarrow A'\}$ for some insufficiency symptom
A' of P wrt I, +Id is the  identifier  associated with a non-root node
of +CPS. Letting the goal associated with  the node  be (A,B), and the
father     of     the     node     be     identified     with     Id'.
icpst_divisor(CPS,Id,CPSa,Sa,CPS1)  succeeds   with  -CPSa   being  the
I-complete  partial  SLDNF tree for $P\cup\{\leftarrow A\}$, -Sa being
the set  of  computed  answers for  A,  and -CPS1  being  obtained  by
replacing  the  subtree  rooted  at  Id'  with  a  tree  obtained from
substituting node Id with a set  of  trees  containing  an  I-complete
partial SLDNF trees for $P\cup\{\leftarrow B\theta\} for each $\theta$
in -Sa.

*/


icpst_divisor(t(Id,G,_,_,CHN),Ida,CPSa,Sa,CPS2) :-
    forest_containingTree(CHN,Ida,CHN_0,t(Id0,G0,Ht0,Wt0,CHN0),CHN_1),
    (  
        Id0 = Ida 
    ->
        icpst_divisor_0(t(Id0,G0,Ht0,Wt0,CHN0),CPSa,Sa,CPSSb),
        append(CHN_0,CPSSb,CHN_0b),append(CHN_0b,CHN_1,CHN2)
    ;
        icpst_divisor(t(Id0,G0,Ht0,Wt0,CHN0),Ida,CPSa,Sa,CPS20),
        append(CHN_0,[CPS20|CHN_1],CHN2)        
    ),
    ordTree_constructor(Id,G,CHN2,CPS2).    

/*

icpst_divisor_0(+CPSab,-CPSa,-Sa,-CPSSb) +CPSab is an I-complete
partial SLDNF tree for $P\cup\{\leftarrow (A,B)\}$.
icpst_divisor_0(CPSab,CPSa,Sa,CPSSb) succeeds with CPSa being an
I-complete partial SLDNF tree for $P\cup\{\leftarrow A\}$, -Sa being
the set of computed answers for A, and -CPSSb containing an I-complete
partial SLDNF tree for $P\cup\{\leftarrow B\theta\}$ for each $\theta$
in -Sa.

*/

icpst_divisor_0(t(Id,W/Sub,Ht,Wt,CHN),Ta,Sa,Fb) :-
    goal_select(W,_A,B),
    number_of_goals(B,Nb),
    icpst_divisor_1(t(Id,W/Sub,Ht,Wt,CHN),Nb,Ta,Sa,Fb).


/*

icpst_divisor_1(+CPSuv,+Nv,-Tu,-Su,-Fv) +CPSuv is an I-complete partial
SLDNF tree for $P\cup\{\leftarrow (U,V)\}$ and +Nv is the number of
literals in +V.  icpst_divisor_1(CPSuv,Nv,Tu,Su,Fv) succeeds with -Tu
being an I-complete partial SLDNF tree for $P\cup\{\leftarrow U\}$,
-Su being the set of computed answers for U and -Fv containing an
I-complete partial SLDNF tree for $P\cup\{\leftarrow V\theta\} for
each compted answer for U in -Su.

*/

icpst_divisor_1(t(Id,W/Sub,Ht,Wt,CHN),Nv,Tu,Su,Fv) :-
    number_of_goals(W, N),
    (
        N =< Nv 
    -> 
        Tu = t(Id,true/Sub,0,1,[]),
        Su = [Sub],
        Fv = [t(Id,W/Sub,Ht,Wt,CHN)]
    ;
        Nu is N - Nv,
        first_goals(W,Nu,Wu),
        icpst_divisor_1_list(CHN,Nv,Fu,Su,Fv),
        ordTree_constructor(Id,Wu/Sub,Fu,Tu)
    ).

icpst_divisor_1_list([],_,[],[],[]).
icpst_divisor_1_list([T|F],Nv,[Tu|Fu],Su,Fv) :-
    icpst_divisor_1(T,Nv,Tu,Su1,Fv1),
    icpst_divisor_1_list(F,Nv,Fu,Su2,Fv2),
    append(Fv1,Fv2,Fv),
    append(Su1,Su2,Su). 

/* 

icpst_transformer1(T1,T2)

*/

icpst_transformer1(t(Id,W/Sub,_,_,CHN),T2) :-
    icpsf_transformer1(CHN,CHN2),
    ordTree_constructor(Id,W/Sub,CHN2,T2).

icpst_transformer11(t(Id,W/Sub,Ht,Wt,CHN),Fo) :- 
    goal_select(W,L,_),
    (   faithful(L)
     -> 
        icpst_divisor_0(t(Id,W/Sub,Ht,Wt,CHN),_,_,F1),
        icpsf_transformer1(F1,Fo)
     ;  
        icpsf_transformer1(CHN,F1),
        ordTree_constructor(Id,W/Sub,F1,To),
        Fo=[To]
    ).

icpsf_transformer1([],[]).
icpsf_transformer1([Ti|Fi],Fo) :-
    icpst_transformer11(Ti,Fo1),
    icpsf_transformer1(Fi,Fo2),
    append(Fo1,Fo2,Fo).

/* icpst_transformer2(+T1,-T2) */

icpst_transformer2(t(Id,W/Sub,_,_,CHN),T2) :-
    icpsf_transformer2(CHN,CHN2),
    ordTree_constructor(Id,W/Sub,CHN2,T2).

/* 

icpst_transformer21(+T,-Tlst) Let T=t(Id,G,H,W,CHN) G=(L,R).
If correct(L) then Tlst is the merge of the lists resulting from
applying icpst_transformer21 to each tree in CHN. Otherwise
Tlst contains a single tree T1 constructed as follows. T1 has the same
Id and label G and the children of T1 is the merge of the lists
resulting from applying icpst_transformer21 to each tree in
CHN.

*/

icpst_transformer21(t(Id,W/Sub,_,_,CHN),Tlst) :- 
    icpsf_transformer2(CHN,CHN2),
    goal_select(W,L,_),
    (   correct(L)
     -> 
        Tlst=CHN2
     ;  
        ordTree_constructor(Id,W/Sub,CHN2,T),
        Tlst=[T]
    ).

icpsf_transformer2([],[]).
icpsf_transformer2([T|Ts],Tlst) :-
    icpst_transformer21(T,Ts1),
    icpsf_transformer2(Ts,Ts2),
    append(Ts1,Ts2,Tlst).

/*number_of_goals(+ConjunctionOfGoals,-NumberOfGoals)   -NumberOfGoals
is the number of +ConjunctionOfGoals.*/

number_of_goals(true,0) :-
    !.
number_of_goals((A,B),N) :-
    !,
    number_of_goals(A,Na),
    number_of_goals(B,Nb),
    N is Na+Nb.
number_of_goals(L,1) :-
    \+ L = (_,_),
    \+ L = true.

/*first_goals(+ConjunctionOfGoals,+NumberOfGoals,-FirstGoals)
-FirstGoals is a prefix of +ConjunctionOfGoals and the number of goals
in -FirstGoals is +NumberOfGoals.*/

first_goals(_W,0,true) :-
    !.
first_goals(L,1,L) :-
    \+ L = (_,_).
first_goals((A,B),N,W) :-
    number_of_goals(A,Na),
    (
        N == Na
    -> 
        W = A
    ;
        N >= Na
    -> 
        N1 is N-Na,
        first_goals(B,N1,W1),
        W = (A,W1)
    ;
        first_goals(A,N,W)        
    ).


/* 

aet(+P,?A,-S)  is a Prolog meta-interpreter.  aet(P,A,S)  solves
+A with +P using depth-first and left-to-right strategy within a depth
  bound.   aet(P,A,S)  succeeds  with  -S  being
true(Sub)  until a run-time error is detected. If a run-time  error is
detected,  aet(P,A,S)  succeeds  with -S  being  Trace
where Cat indicates  the category of the error while Trace records the
history of the  exectution of A  until the point at which the error is
detected.

*/

aet(P,A,S) :-
    read_parameter_value(depth_bound,D),
    vars(A,Vs),
    aet(A,(('$Goal'/0,1),1),P,Vs,[],D,0,_,S),
                         % 0 labels the root
    (
       \+ S = true(_),
       !  % prevent backtracking from taking place.
    ;
       true
    ).

/*aet(+A,+P,+Vs,+CSub,+D,+Idr,-Idn,-S) is a depth bound
meta-interpreter for normal logic programs.
aet(A,P,Vs,CSub,D,Idr,Idn,S) solves +A\Input with program +P using
depth-first and left-to-right strategy within depth bound D.
aet(A,P,Vs,CSub,D,R,Idr,Idn,S) succeeds if there is a SLDNF
refutation for $P\cup\{\leftarrow A\Input\}$ within depth bound D and
with -S being true(SubOut) and SubOut being output substitution.
aet(A,P,Vs,CSub,D,Idr,Idn,S) also succeeds if a run-time error is
detected with S being Trace where -Trace the execution
history.*/

aet(_A,(Cid,Gid),_P,_Vs,CS,0,Min,Min,t(Min,((Cid,Gid),CS),0,1,[])) :- 
    !.
aet(\+(A),(Cid,Gid),P,Vs,CS,D,Min,Max,R) :-
%    D>0,
    !,
    D1 is D-1,
    Min0 is Min + 1,
    (   
	aet(A,(Cid,Gid),P,Vs,CS,D1,Min0,Max0,Ra)
    ->
 	(
	    Ra = true(RS)
	->
            sub_apply(CS,A,A1),
	    sub_apply(RS,A,A2),
	    (
		\+ variant(A1,A2)
	    ->
		Gid1 is Gid+1, Max = Min,
		R  = t(Min,((Cid,Gid1),RS),0,1,[])
	    )
	;   
            R = Ra, Max = Max0
    	)		
    ;
	R=true(CS), Max = Min
    ).
    
aet((A,B),(Cid,Gid),P,Vs,CS,D,Min,Max,R) :- 
%    D>0,
    !,
    aet(A,(Cid,Gid),P,Vs,CS,D,Min,MaxA,Ra),
    (
    	Ra = true(RSa)
    -> 
	MinB is MaxA + 1,
	( Gid1 is Gid+1 ),
       	aet(B,(Cid,Gid1),P,Vs,RSa,D,MinB,MaxB,Rb),
       	(
            Rb = true(RSb)
	->
            R  = true(RSb), Max = Min
       	;
            (
            	( Rb = [_|_] -> Rb0 = Rb ; Rb0 = [Rb] ),
		sub_apply(RSa,A,A1),
                R=[t(Min,A1,0,1,[])|Rb0]
            ),
	    Max = MaxB
       	)
    ; 
    	R = Ra,	Max = MaxA
    ).

aet(A,(_Cid,_Gid),_P,Vs,CS,_D,Min,Min,R) :- 
        % assumes that each bip is used correctly, needs to be modified
    bip(A),
    !,
    sub_apply(CS,A,Goal),
    term_rename(Goal,Renamed),
    call(Renamed),
    mgu(Renamed,Goal,Answer),
    sub_compose(CS,Answer,RSub0),
    sub_restrict(RSub0,Vs,RSub),
    R = true(RSub). 


aet(A,(Cid0,Gid0),P,Vs,CS,D,Min,Max,R) :-
    D1 is D-1,
    sub_apply(CS,A,Goal),
    term_rename(Goal,Renamed),
    functor(A,PN,PA),
    pg_EnumerateCLS(PN/PA,P,Cid,(Head:-Body)),
    mgu(Head,Renamed,Entry0),
    vars((Head:-Body),VHB),
    sub_restrict(Entry0,VHB,Entry),
    MinB is Min + 1,
    aet(Body,((PN/PA,Cid),1),P,VHB,Entry,D1,MinB,MaxB,Rb),
    (
        Rb = true(RSb)
    ->
        sub_apply(RSb,Head,Solved),
        term_rename(Solved,RenamedSolved),
        mgu(RenamedSolved,Goal,RSub0),
        sub_compose(CS,RSub0,RSub1),
        sub_restrict(RSub1,Vs,RSub),
        R = true(RSub),
        Max = Min
    ;
        ( Rb = [_|_] -> Rb0 = Rb ; Rb0 = [Rb] ),
        ordTree_constructor(Min,((Cid0,Gid0),CS),Rb0,R),
        Max = MaxB
    ).


